package games.strategy.triplea.ai.TestAI;

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

import java.util.ArrayList;

import games.strategy.engine.data.GameData;
import games.strategy.engine.data.PlayerID;
import games.strategy.engine.data.Territory;
import games.strategy.engine.data.Unit;
import games.strategy.triplea.attatchments.TerritoryAttachment;
import games.strategy.triplea.attatchments.UnitAttachment;
import games.strategy.triplea.delegate.Matches;
import games.strategy.triplea.delegate.dataObjects.PlaceableUnits;
import games.strategy.triplea.delegate.remote.IAbstractPlaceDelegate;

/**
 * TestPlaceAI
 * 
 * @author A.S.
 * @since 2015
 */

public class TestPlaceAI 
{
	private final static SimpleLogger logger = new SimpleLogger();
	
	private IAbstractPlaceDelegate placeDelegate;
	private PlayerID player;
	private GameData gameData;
	private ArrayList<TestPlaceContainer> ownFactoriesList;
	
	private int alignmentOfAI;
	
	private boolean stopPlace;
	
	public TestPlaceAI(final boolean placeForBid, final IAbstractPlaceDelegate placeDelegate, final GameData data, final PlayerID player, int alignmentOfAI)
	{
		this.placeDelegate = placeDelegate;
		this.alignmentOfAI = alignmentOfAI;
		this.player = player;
		
		gameData = data;
		
		ownFactoriesList = new ArrayList<TestPlaceContainer>();
		
		stopPlace = false;
	}
	
	public void handlePlace()
	{	 
		/*
		 * Relevant area for the counting of own and enemy capitals/victory-cities/units
		 *
		 * The set distances are important for the formula in computePlace() and hence playing strength of TestPlaceAI
		 */
		int shortTerritoryDistance = 2;
		int largeTerritoryDistance = 4;
		
		/*
		 * Step 1:
		 *
		 * Enlist own factories and set their properties
		 */
		for (Territory territoryWithOwnFactory:gameData.getMap().getTerritories())
		{
			if (territoryWithOwnFactory.getOwner().equals(player) && territoryWithOwnFactory.getUnits().someMatch(Matches.UnitCanProduceUnits))
			{
				TestPlaceContainer currentFactory = new TestPlaceContainer();
				
				currentFactory.setTerritoryWithOwnFactory(territoryWithOwnFactory);			
				
				// Check for adjacent sea territory
				for (Territory neighbour:gameData.getMap().getNeighbors(territoryWithOwnFactory))
				{
					if (neighbour.isWater())
					{
						currentFactory.setAdjacentSeaTerritory(neighbour);
					}
				}
				
				// Set count of capital-, victory-cities and units within distance to the current factory (= territory)		
				int ownCapitalInZone = 0;
				int ownVictoryCitiesInZone = 0;
				int ownUnitsInZone = 0;
				int enemyUnitsInZone = 0;
				int enemyCapitolInZone = 0;
				int enemyVictoryCitiesInZone = 0;
				
				for (Territory territoryInZone:gameData.getMap().getTerritories())
				{
					if (gameData.getMap().getDistance(territoryInZone, territoryWithOwnFactory) <= shortTerritoryDistance && TerritoryAttachment.get(territoryInZone) != null)
					{
						if (TerritoryAttachment.get(territoryInZone).isCapital() && territoryInZone.getOwner().equals(player))
						{	
							ownCapitalInZone++;
						}
						
						if (TerritoryAttachment.get(territoryInZone).getVictoryCity() == 1 && territoryInZone.getOwner().equals(player))
						{
							ownVictoryCitiesInZone++;
						}
						
						ownUnitsInZone += territoryInZone.getUnits().getUnitCount(player);
						enemyUnitsInZone += (territoryInZone.getUnits().getUnitCount()-territoryInZone.getUnits().getUnitCount(player));

						if (gameData.getMap().getDistance(territoryInZone, territoryWithOwnFactory) <= largeTerritoryDistance && TerritoryAttachment.get(territoryInZone) != null) 
						{
							if (TerritoryAttachment.get(territoryInZone).isCapital() && !(territoryInZone.getOwner().equals(player)))
							{
								enemyCapitolInZone++;
							}
							
							if (TerritoryAttachment.get(territoryInZone).getVictoryCity() == 1 && !(territoryInZone.getOwner().equals(player)))
							{
								enemyVictoryCitiesInZone++;
							}
						}
					}
				}
				
				currentFactory.setOwnCapitalInZone(ownCapitalInZone);
				currentFactory.setOwnVictoryCitiesInZone(ownVictoryCitiesInZone);
				currentFactory.setOwnUnitsInZone(ownUnitsInZone);
				currentFactory.setEnemyCapitalInZone(enemyCapitolInZone);
				currentFactory.setEnemyVictoryCitiesInZone(enemyVictoryCitiesInZone);
				currentFactory.setEnemyUnitsInZone(enemyUnitsInZone);
				
				ownFactoriesList.add(currentFactory);
			}
		}
		
		/* 
		 * Step 2: 
		 * 
		 * Calculate the place-attractivity of the available factories and purchased unit types (see "TestPurchaseAI") according to the current alignment of the TestAI player and place units
		 */		
		while (player.getUnits().getUnits().size() != 0 && !stopPlace)
		{
			if (alignmentOfAI == 1) // defensive
			{
				computePlace(2, 1.5, 1, 1, 1, 2);
			}
			else if (alignmentOfAI == 2) // normal
			{
				computePlace(1.5, 1.3, 1.2, 1.1, 1, 1);
			}
			else if (alignmentOfAI == 3) // offensive
			{
				computePlace(1, 1, 2, 2, 2, 1);
			}
			else
			{
				logger.logMessage("Place stopped: no valid alignment of TestAI player");
			}
		}
	}
	
	/*
	 * Selects a factory where and a single unit to be placed
	 */
	private void computePlace(double ownCapitalInZoneWeight, double ownVictoryCitiesInZoneWeight, double enemyCapitalInZoneWeight, double enemyVictoryCitiesInZoneWeight, double ownUnitsInZoneWeight, double enemyUnitsInZoneWeight)
	{		
		/*
		 * Step 1:
		 * 
		 * Calculate the place-attractivity of each factory
		 * 
		 * I.e., which own factory (and surrounding area) would profit most from new units in the current situation
		 */
		for (TestPlaceContainer currentFactory:ownFactoriesList)
		{
			double placeAttractivityUnit = 0;
			
			if (currentFactory.getEnemyUnitsInZone() != 0)
				currentFactory.setUnitRatio((double)currentFactory.getOwnUnitsInZone()/currentFactory.getEnemyUnitsInZone());
			} else {
				currentFactory.setUnitRatio(currentFactory.getOwnUnitsInZone());
			}
			
			// Calculate the unit ratio (own to enemy) in the relevant area (see step 1 in handlePlace()) around the territory of the factory
			placeAttractivityUnit = currentFactory.getOwnCapitalInZone()*ownCapitalInZoneWeight+
					currentFactory.getOwnVictoryCitiesInZone()*ownVictoryCitiesInZoneWeight+
					currentFactory.getOwnUnitsInZone()*ownUnitsInZoneWeight+
					currentFactory.getEnemyCapitalInZone()*enemyCapitalInZoneWeight+
					currentFactory.getEnemyVictoryCitiesInZone()*enemyVictoryCitiesInZoneWeight+
					currentFactory.getEnemyUnitsInZone()*enemyUnitsInZoneWeight;
			
			currentFactory.setPlaceAttractivityUnit(placeAttractivityUnit);
		}
		
		// logger.logMessage("ownFactoriesListSize: " + ownFactoriesList.size() + " unitsToPlace: " + player.getUnits().getUnits().size()); // Part of test log: "Placement of units"
		
		// Check, if there is at least one factory available for placement
		if (ownFactoriesList.size() != 0)
		{
			/*
			 * Step 2:
			 * 
			 * Select the factory (territory), where a unit should be placed in this run
			 */
			double dummyPlaceAttractivityUnit = 0;
			int listIndex = 0;
			
			for (TestPlaceContainer currentFactory:ownFactoriesList)
			{
				if (currentFactory.getPlaceAttractivityUnit() > dummyPlaceAttractivityUnit)
				{
					dummyPlaceAttractivityUnit = currentFactory.getPlaceAttractivityUnit();
					listIndex = ownFactoriesList.indexOf(currentFactory);
				}
			}
	
			Territory territoryToPlaceAt = ownFactoriesList.get(listIndex).getTerritoryWithOwnFactory();
						
			/*
			 * Step 3:
			 * 
			 * Select the unit, that should be placed on the chosen territory
			 */
			final ArrayList<Unit> unitToPlace = new ArrayList<Unit>();
			
			// Selects the most appropriate unit based on the unit ratio in the relevant area for the territory
			unitToPlace.add(selectUnitToPlace(ownFactoriesList.get(listIndex).getUnitRatio()));
			
			if (unitToPlace.get(0) == null)
			{
				// logger.logMessage("selectUnitToPlace returns null"); // Part of test log: "Placement of units"
				return;
			}
			
			/* 
			 * Step 4:
			 * 
			 * Check, if the selected unit can be placed on the chosen territory
			 * 
			 * If the selected unit can not be placed on the chosen territory, the next unit shall be selected.
			 * If no unit can be placed on the chosen territory, the factory shall be removed from the list.
			 *
			 * TODO: overall, it could be more reasonable to loop through the purchased units, and select the most appropriate factory, than vice versa (as we do now). I.e. one question is, if 
			 * we shall place as many purchased units we can in this turn, or it is more important to strengthen certain factories (territories, areas). Of course, an ideal approach
			 * would always find an optimised solution for both considerations.
			 */
			
			// Check for placement of marine units
			if (Matches.UnitIsSea.match(unitToPlace.get(0)) && ownFactoriesList.get(listIndex).getAdjacentSeaTerritory() == null)
			{
				return;
			}
			else if (Matches.UnitIsSea.match(unitToPlace.get(0)) && ownFactoriesList.get(listIndex).getAdjacentSeaTerritory() != null)
			{
				territoryToPlaceAt = ownFactoriesList.get(listIndex).getAdjacentSeaTerritory();
			}
			
			final PlaceableUnits placeableUnits = placeDelegate.getPlaceableUnits(unitToPlace, territoryToPlaceAt);
			
			// Check for placement-limit and other errors (e.g. "territory was conquered")
			if (placeableUnits.getErrorMessage() != null || TerritoryAttachment.get(ownFactoriesList.get(listIndex).getTerritoryWithOwnFactory()).getUnitProduction() < ownFactoriesList.get(listIndex).getUnitsPlacedInTerritory())
			{
				//logger.logMessage("Unit can not be placed on territory: " + placeableUnits.getErrorMessage()); // Part of test log: "Placement of units"
				
				ownFactoriesList.remove(listIndex); // remove factory
	
				return;
			}
			
			//logger.logMessage("Unit will be placed: " + placeableUnits.getUnits().iterator().next() + " on territory " + territoryToPlaceAt.getName());
		
			/*
			 * Step 5:
			 * 
			 * Perform place for the chosen unit and factory (territory)
			 */
			
			 /*
			  * Manually increase unit count by 1
			  *  
			  * TODO: there should be a way to do this "automatically", such that the unit is added for the unit count of other factories too -> e.g., if we manually add one
			  * unit to "Caucasus", it would be in the range of, but not counted for the factory in "Russia" -> i.e., at least we could call "setOwnUnitsInZone" for all affected factories
			  */
			ownFactoriesList.get(listIndex).setOwnUnitsInZone(ownFactoriesList.get(listIndex).getOwnUnitsInZone()+1);
			
			//logger.logMessage("unitToPlace: " + unitToPlace.get(0).toString() + " territoryToPlaceAt: " + territoryToPlaceAt.getName().toString()); // Part of test log: "Placement of units"
			
			ownFactoriesList.get(listIndex).setUnitsPlacedInTerritory();
			
			doPlace(unitToPlace, territoryToPlaceAt);
		}
		else
		{
			stopPlace = true;
		}
	}
	
	/*
	 * Places unit with placeDelegate
	 */
	private void doPlace(ArrayList<Unit> unitToPlace, Territory territoryToPlaceAt)
	{	
		placeDelegate.placeUnits(new ArrayList<Unit>(unitToPlace), territoryToPlaceAt);
	}
	
	/*
	 * Selects the "most appropriate" unit of the available units based on the specified unit ratio
	 *
	 * E.g., if the enemy has significantly more units, a defensive unit is prefered.
	 * 
	 * TODO: find an more elaborated approach -> e.g., the effectiveness of own and enemy unit types (against each other) could be considered, and not only their absolute count
	 */
	private Unit selectUnitToPlace(double ratioOfOwnUnitsToEnemies)
	{
		double dummyValue = 0;
		double dummyValueOverall = 0;
		Unit appropriateUnitRegardingTerritorySituation = null;
		
		for (Unit currentUnit:player.getUnits().getUnits())
		{
			if (ratioOfOwnUnitsToEnemies < 0.50) // formula for preferential treatment of defensive units
			{
				dummyValue = UnitAttachment.get(currentUnit.getUnitType()).getAttack(player)*1 +
						UnitAttachment.get(currentUnit.getUnitType()).getDefense(player)*2 +
						UnitAttachment.get(currentUnit.getUnitType()).getMovement(player)*1;
						
				if (dummyValue > dummyValueOverall)
				{
					dummyValueOverall = dummyValue;
					appropriateUnitRegardingTerritorySituation = currentUnit;
				}
			}
			else if (ratioOfOwnUnitsToEnemies >= 0.50 && ratioOfOwnUnitsToEnemies < 1) // formula for preferential treatment of normal units
			{
				dummyValue = UnitAttachment.get(currentUnit.getUnitType()).getAttack(player)*1 +
						UnitAttachment.get(currentUnit.getUnitType()).getDefense(player)*1 +
						UnitAttachment.get(currentUnit.getUnitType()).getMovement(player)*1;
						
				if(dummyValue > dummyValueOverall)
				{
					dummyValueOverall = dummyValue;
					appropriateUnitRegardingTerritorySituation = currentUnit;
				}
			}
			else if (ratioOfOwnUnitsToEnemies >= 1) // formula for preferential treatment of offensive units
			{
				dummyValue = UnitAttachment.get(currentUnit.getUnitType()).getAttack(player)*2 +
						UnitAttachment.get(currentUnit.getUnitType()).getDefense(player)*1 +
						UnitAttachment.get(currentUnit.getUnitType()).getMovement(player)*1;
						
				if (dummyValue > dummyValueOverall)
				{
					dummyValueOverall = dummyValue;
					appropriateUnitRegardingTerritorySituation = currentUnit;
				}
			}
			else
			{
				logger.logMessage("Invalid value for ratioOfOwnUnitsToEnemies in selectUnitToPlace in TestPlaceAI");
			}
		}
		
		//logger.logMessage("Unit selected by selecUnitToPlace: " + appropriateUnitRegardingTerritorySituation.toString() + " " + appropriateUnitRegardingTerritorySituation.getID());
		
		return appropriateUnitRegardingTerritorySituation;
	}
}